"""The view layer parses the request, and dispatches it to an operation."""
import re
from typing import List, Optional, Type

from django.core.exceptions import (
    ImproperlyConfigured,
    PermissionDenied as Django_PermissionDenied,
    SuspiciousOperation,
)
from django.http import HttpResponse
from django.views import View

from gisserver import conf
from gisserver.exceptions import (
    InvalidParameterValue,
    MissingParameterValue,
    OWSException,
    OperationNotSupported,
    PermissionDenied,
)
from gisserver.features import FeatureType, ServiceDescription
from gisserver.operations import base, wfs20

SAFE_VERSION = re.compile(r"\A[0-9.]+\Z")


class GISView(View):
    """The base logic to implement OGC view like WFS.

    Each subclass defines 'accept_operations' with the desired RPC operations.
    """

    #: Define the namespace to use in the XML
    xml_namespace = "http://example.org/gisserver"

    #: Default version to use
    version = "2.0.0"

    #: Supported versions
    accept_versions = ("2.0.0",)

    #: Internal configuration of all available RPC calls.
    accept_operations = {}

    #: Metadata of the service:
    service_description: Optional[ServiceDescription] = None

    def dispatch(self, request, *args, **kwargs):
        """Render proper XML errors for exceptions on all request types."""
        try:
            return super().dispatch(request, *args, **kwargs)
        except Exception as e:
            response = self.handle_exception(e)
            if response is not None:
                return response
            raise

    def handle_exception(self, exc):
        """Transform an exception into a OGC response.
        When nothing is returned, the exception is raised instead.
        """
        if isinstance(exc, Django_PermissionDenied):
            exc = PermissionDenied("typeNames", text=str(exc) or None)
            return HttpResponse(
                exc.as_xml().encode("utf-8"),
                content_type="text/xml; charset=utf-8",
                status=exc.status_code,
                reason=exc.reason,
            )
        elif isinstance(exc, OWSException):
            # Wrap our XML-based exception into a response.
            exc.version = self.version  # Use negotiated version
            return HttpResponse(
                exc.as_xml().encode("utf-8"),
                content_type="text/xml; charset=utf-8",
                status=exc.status_code,
                reason=exc.reason,
            )

    def get(self, request, *args, **kwargs):
        """Entry point to handle HTTP GET requests.

        This parses the 'SERVICE' and 'REQUEST' parameters,
        to call the proper operation.

        All query parameters are handled as case-insensitive.
        """
        # Convert to WFS key-value-pair format.
        self.KVP = {key.upper(): value for key, value in request.GET.items()}
        wfs_method_cls = self.get_operation_class()
        return self.call_operation(wfs_method_cls)

    def get_operation_class(self) -> Type[base.WFSMethod]:
        """Resolve the method that the client wants to call."""
        if not self.accept_operations:
            raise ImproperlyConfigured("View has no operations")

        # The service is always WFS
        service = self._get_required_arg("SERVICE").upper()
        try:
            operations = self.accept_operations[service]
        except KeyError:
            allowed = ", ".join(sorted(self.accept_operations.keys()))
            raise InvalidParameterValue(
                "service",
                f"'{service}' is an invalid service, supported are: {allowed}.",
            ) from None

        # Resolve the operation
        # In mapserver, the operation name is case insensitive.
        operation = self._get_required_arg("REQUEST").upper()
        uc_methods = {name.upper(): method for name, method in operations.items()}

        try:
            return uc_methods[operation]
        except KeyError:
            allowed = ", ".join(operations.keys())
            raise OperationNotSupported(
                "request",
                f"'{operation.lower()}' is not implemented, supported are: {allowed}.",
            ) from None

    def call_operation(self, wfs_method_cls: Type[base.WFSMethod]):
        """Call the resolved method."""
        wfs_method = wfs_method_cls(self)
        param_values = wfs_method.parse_request(self.KVP)
        return wfs_method(**param_values)  # goes into __call__()

    def _get_required_arg(self, argname):
        try:
            return self.KVP[argname]
        except KeyError:
            raise MissingParameterValue(argname.lower()) from None

    def get_service_description(self, service: str) -> ServiceDescription:
        """Provide the (dynamically generated) service description."""
        return self.service_description or ServiceDescription(title="Unnamed")

    def set_version(self, version):
        """Enforce a particular version based on the request."""
        if not SAFE_VERSION.match(version):
            # Make really sure we didn't mess things up, as this causes file includes.
            raise SuspiciousOperation("Invalid/insecure version number parsed")

        # Enforce the requested version
        self.version = version

    @property
    def server_url(self):
        """Expose the server URLs for all operations to read."""
        return self.request.build_absolute_uri(self.request.path)


class WFSView(GISView):
    """A view for a single WFS server.

    This view exposes multiple dataset,
    each containing a single Django model (mapped to feature types in WFS).

    This class can be used by subclassing it, and redefining ``feature_types``
    or ``get_feature_types()``.
    """

    #: Maximum number of features to return
    max_page_size = conf.GISSERVER_DEFAULT_MAX_PAGE_SIZE

    #: Define the features (=tables) in this dataset.
    feature_types: List[FeatureType] = []

    #: Internal configuration of all available RPC calls.
    accept_operations = {
        "WFS": {
            "GetCapabilities": wfs20.GetCapabilities,
            "DescribeFeatureType": wfs20.DescribeFeatureType,
            "GetFeature": wfs20.GetFeature,
            "GetPropertyValue": wfs20.GetPropertyValue,
            "ListStoredQueries": wfs20.ListStoredQueries,
            "DescribeStoredQueries": wfs20.DescribeStoredQueries,
        }
    }

    #: Metadata of the capabilities:
    wfs_service_constraints = {
        "ImplementsBasicWFS": False,  # only simple WFS
        "ImplementsTransactionalWFS": False,  # only simple WFS
        "ImplementsLockingWFS": False,  # only simple WFS
        "KVPEncoding": True,  # HTTP GET support
        "XMLEncoding": False,  # HTTP POST requests
        "SOAPEncoding": False,  # no SOAP requests
        "ImplementsInheritance": False,
        "ImplementsRemoteResolve": False,
        "ImplementsResultPaging": True,  # missing next/previous for START / COUNT
        "ImplementsStandardJoins": False,  # returns records as wfs:Tuple in GetFeature
        "ImplementsSpatialJoins": False,
        "ImplementsTemporalJoins": False,
        "ImplementsFeatureVersioning": False,
        "ManageStoredQueries": False,
        #
        # Mentioned as operation constraint in WFS spec:
        "PagingIsTransactionSafe": False,
    }

    #: Metadata of filtering capabilities
    wfs_filter_capabilities = {
        "ImplementsQuery": True,  # <fes:AbstractQueryElement> needed for WFS simple
        "ImplementsAdHocQuery": True,  # <fes:AbstractAdhocQueryElement> needed for WFS simple
        "ImplementsFunctions": True,  # <fes:Function> support
        "ImplementsResourceId": True,  # <fes:ResourceId> support
        "ImplementsMinStandardFilter": True,  # <fes:PropertyIs...> support
        "ImplementsStandardFilter": True,  # <fes:And>, <fes:Or> and advanced functions
        "ImplementsMinSpatialFilter": True,  # <fes:BBOX>
        "ImplementsSpatialFilter": True,  # Other spatial functions
        "ImplementsMinTemporalFilter": False,  # mapserver: true (During)
        "ImplementsTemporalFilter": False,
        "ImplementsVersionNav": False,  # <fes:ResourceId version="..">
        "ImplementsSorting": True,  # SORTBY parameter
        "ImplementsExtendedOperators": False,
        "ImplementsMinimumXPath": True,  # incomplete like mapserver, needed for cite
        "ImplementsSchemaElementFunc": False,
    }

    def get_feature_types(self) -> List[FeatureType]:
        """Return all available feature types this server exposes"""
        return self.feature_types



"""These classes map to the FES 2.0 specification for expressions.
The class names are identical to those in the FES spec.
"""
import operator
from dataclasses import dataclass
from datetime import date, datetime
from decimal import Decimal as D
from django.utils.functional import cached_property
from typing import List, Optional, Tuple, Union
from xml.etree.ElementTree import Element

from django.contrib.gis.geos import GEOSGeometry
from django.db.models import F, Func, Q, Value
from django.db.models.expressions import Combinable

from gisserver.parsers.base import FES20, BaseNode, TagNameEnum, tag_registry
from gisserver.parsers.fes20.functions import function_registry
from gisserver.parsers.utils import (
    auto_cast,
    expect_tag,
    get_attribute,
    xsd_cast,
)

NoneType = type(None)

RhsTypes = Union[
    Combinable, Func, Q, GEOSGeometry, bool, int, str, date, datetime, tuple
]


class BinaryOperatorType(TagNameEnum):
    """FES 1.0 Arithmetic operators.

    This are no longer part of the FES 2.0 spec, but clients (like QGis)
    still assume the server supports these. Hence these need to be included.
    """

    Add = operator.add
    Sub = operator.sub
    Mul = operator.mul
    Div = operator.truediv


class Expression(BaseNode):
    """Abstract base class, as defined by FES spec."""

    xml_ns = FES20

    def build_lhs(self, compiler) -> str:
        """Get the expression as the left-hand-side of the equation.

        This typically returns the expression as a 'field name' which can be
        used in the Django QuerySet.filter(name=...) syntax. When the
        expression is actually a Function/Literal, this should generate the
        name using a queryset annotation.
        """
        value = _make_combinable(self.build_rhs(compiler))
        return compiler.add_annotation(value)

    def build_rhs(self, compiler) -> RhsTypes:
        """Get the expression as the right-hand-side of the equation.

        Typically, this can return the exact value.
        """
        raise NotImplementedError(f"{self.__class__.__name__}.build_rhs()")


@dataclass
@tag_registry.register("Literal")
class Literal(Expression):
    """The <fes:Literal> element that holds a literal value"""

    value: Union[int, str, date, D, datetime, NoneType]  # officially <xsd:any>
    type: Optional[str] = None

    def __str__(self):
        return self.value

    @classmethod
    @expect_tag(FES20, "Literal", leaf=True)
    def from_xml(cls, element: Element):
        type = element.get("type")
        value = element.text

        if type:
            # Cast the value based on the given xsd:QName
            value = xsd_cast(value, type)
        else:
            # Make sure gt / datetime comparisons work out of the box.
            value = auto_cast(value)

        return cls(value=value, type=type)

    def build_lhs(self, compiler) -> str:
        """Alias the value when it's used in the left-hand-side.

        By aliasing the value using an annotation,
        it can be queried like a regular field name.
        """
        return compiler.add_annotation(Value(self.value))

    def build_rhs(self, compiler) -> Union[Combinable, Q, str]:
        """Return the value when it's used in the right-hand-side"""
        return self.value


@dataclass
@tag_registry.register("ValueReference")
@tag_registry.register("PropertyName")  # FES 1.0 name, some clients still use this.
class ValueReference(Expression):
    """The <fes:ValueReference> element that holds an XPath string.
    In the fes XSD, this is declared as a subclass of xsd:string.

    The old WFS1/FES1 "PropertyName" is allowed as an alias.
    Various clients still send this, and mapserver/geoserver support this.
    """

    xpath: str

    def __str__(self):
        return self.xpath

    @classmethod
    @expect_tag(FES20, "ValueReference", "PropertyName", leaf=True)
    def from_xml(cls, element: Element):
        return cls(xpath=element.text)

    def build_lhs(self, compiler) -> str:
        """Optimized LHS: there is no need to alias a field lookup through an annotation."""
        field, extra_q = self.parse_xpath(compiler.feature_type)
        if extra_q:
            compiler.add_extra_lookup(extra_q)
        return field

    def build_rhs(self, compiler) -> RhsTypes:
        """Return the value as F-expression"""
        return F(self.build_lhs(compiler))

    def parse_xpath(self, feature_type=None) -> Tuple[str, Optional[Q]]:
        """Convert the XPath into a the required ORM query elements."""
        if feature_type is not None:
            # Can resolve against XSD paths, find the correct DB field name
            match = feature_type.resolve_element(self.xpath)
            return match.orm_path, match.orm_filters
        else:
            # Only used by unit testing (when feature_type is not given).
            parts = [word.strip() for word in self.xpath.split("/")]
            return "__".join(parts), None

    @cached_property
    def element_name(self):
        """Tell which element this reference points to."""
        try:
            pos = self.xpath.rindex("/")
        except ValueError:
            return self.xpath
        else:
            return self.xpath[pos + 1 :]


@dataclass
@tag_registry.register("Function")
class Function(Expression):
    """The <fes:Function name="..."> element."""

    name: str  # scoped name
    arguments: List[Expression]  # xsd:element ref="fes20:expression"

    @classmethod
    @expect_tag(FES20, "Function")
    def from_xml(cls, element: Element):
        return cls(
            name=get_attribute(element, "name"),
            arguments=[Expression.from_child_xml(child) for child in element],
        )

    def build_rhs(self, compiler) -> RhsTypes:
        """Build the SQL function object"""
        db_function = function_registry.resolve_function(self.name)
        args = [arg.build_rhs(compiler) for arg in self.arguments]
        return db_function.build_query(*args)


@dataclass
@tag_registry.register_names(BinaryOperatorType)
class BinaryOperator(Expression):
    """Support for FES 1.0 arithmetic operators.

    This are no longer part of the FES 2.0 spec, but clients (like QGis)
    still assume the server supports these. Hence these need to be included.
    """

    _operatorType: BinaryOperatorType
    expression: Tuple[Expression, Expression]

    @classmethod
    def from_xml(cls, element: Element):
        return cls(
            _operatorType=BinaryOperatorType.from_xml(element),
            expression=(
                Expression.from_child_xml(element[0]),
                Expression.from_child_xml(element[1]),
            ),
        )

    def build_rhs(self, compiler) -> RhsTypes:
        value1 = _make_combinable(self.expression[0].build_rhs(compiler))
        value2 = _make_combinable(self.expression[1].build_rhs(compiler))
        return self._operatorType.value(value1, value2)


def _make_combinable(value) -> Union[Combinable, Q]:
    """Make sure the scalar value is wrapped inside a compilable object"""
    if isinstance(value, (Combinable, Q)):
        return value
    else:
        return Value(value)  # e.g. str, or GEOSGeometry



from django.db.models import QuerySet
from typing import Dict, List, Optional, Tuple

from gisserver.exceptions import InvalidParameterValue
from gisserver.features import FeatureType
from gisserver.output import FeatureCollection, SimpleFeatureCollection
from gisserver.parsers import fes20
from gisserver.types import strip_namespace_prefix


class QueryExpression:
    """WFS base class for all queries.
    This object type is defined in the WFS spec (as <fes:AbstractQueryExpression>).

    The subclasses can override the following logic:

    * :meth:`get_type_names` defines which types this query applies to.
    * :meth:`compile_query` defines how to filter the queryset.

    For full control, these methods can also be overwritten instead:

    * :meth:`get_queryset` defines the full results.
    * :meth:`get_hits` to return the collection for RESULTTYPE=hits.
    * :meth:`get_results` to return the collection for RESULTTYPE=results
    """

    handle = ""
    value_reference = None

    def bind(
        self,
        all_feature_types: Dict[str, FeatureType],
        value_reference: Optional[fes20.ValueReference],
    ):
        """Bind the query to presentation-layer logic"""
        self.all_feature_types = all_feature_types
        self.value_reference = value_reference

    def check_permissions(self, request):
        """Verify whether the user has access to view these data sources"""
        for feature_type in self.get_type_names():
            feature_type.check_permissions(request)

    def resolve_type_name(self, type_name, locator="") -> FeatureType:
        """Find the feature type for a given name.
        This is an utility that cusstom subclasses can use.
        """
        # Strip the namespace prefix. The Python ElementTree parser does
        # not expose the used namespace prefixes, so text-values can't be
        # mapped against it. As we expose just one namespace, just strip it.
        type_name = strip_namespace_prefix(type_name)

        try:
            return self.all_feature_types[type_name]
        except KeyError:
            raise InvalidParameterValue(
                "typename", f"Typename '{type_name}' doesn't exist in this server."
            ) from None

    def get_hits(self) -> FeatureCollection:
        """Run the query, return the number of hits only.

        Override this method in case you need full control over the response data.
        Otherwise, override :meth:`compile_query` or :meth:`get_queryset`.
        """
        querysets = self.get_querysets()
        return FeatureCollection(
            results=[
                # Include empty feature collections,
                # so the selected feature types are still known.
                SimpleFeatureCollection(
                    feature_type=ft, queryset=qs.none(), start=0, stop=0
                )
                for ft, qs in querysets
            ],
            number_matched=sum([qs.count() for ft, qs in querysets]),
        )

    def get_results(self, start_index=0, count=100) -> FeatureCollection:
        """Run the query, return the full paginated results.

        Override this method in case you need full control over the response data.
        Otherwise, override :meth:`compile_query` or :meth:`get_queryset`.
        """
        stop = start_index + count

        # The querysets are not executed yet, until the output is reading them.
        querysets = self.get_querysets()
        return FeatureCollection(
            results=[
                SimpleFeatureCollection(feature_type, qs, start=start_index, stop=stop)
                for feature_type, qs in querysets
            ]
        )

    def get_querysets(self) -> List[Tuple[FeatureType, QuerySet]]:
        """Construct the querysets that return the database results."""
        results = []
        for feature_type in self.get_type_names():
            queryset = self.get_queryset(feature_type)
            results.append((feature_type, queryset))

        return results

    def get_queryset(self, feature_type: FeatureType) -> QuerySet:
        """Generate the queryset for the specific feature type.

        This method can be overwritten in subclasses to define the returned data.
        However, consider overwriting :meth:`compile_query` instead of simple data.
        """
        queryset = feature_type.get_queryset()

        # Apply filters
        compiler = self.compile_query(feature_type)

        if self.value_reference is not None:
            if feature_type.resolve_element(self.value_reference.xpath) is None:
                raise InvalidParameterValue(
                    "valueReference",
                    f"Field '{self.value_reference.xpath}' does not exist.",
                )

            # For GetPropertyValue, adjust the query so only that value is requested.
            # This makes sure XPath attribute selectors are already handled by the
            # database query, instead of being a presentation-layer handling.
            field = compiler.add_value_reference(self.value_reference)
            queryset = compiler.filter_queryset(queryset, feature_type=feature_type)
            return queryset.values("pk", member=field)
        else:
            return compiler.filter_queryset(queryset, feature_type=feature_type)

    def get_type_names(self) -> List[FeatureType]:
        """Tell which type names this query applies to.

        This method needs to be defined in subclasses.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__}.get_type_names() should be implemented."
        )

    def compile_query(self, feature_type: FeatureType) -> fes20.CompiledQuery:
        """Define the compiled query that filters the queryset.

        Subclasses need to define this method, unless
        :meth:`get_queryset` is completely overwritten.
        """
        raise NotImplementedError()



"""Generic support for various GML versions.

These functions locate GML objects, and redirect to the proper parser.
"""
from typing import List, Union
from xml.etree.ElementTree import Element

from defusedxml.ElementTree import fromstring
from gisserver.exceptions import ExternalValueError
from gisserver.parsers.base import tag_registry

from .base import AbstractGeometry, GM_Envelope, GM_Object, TM_Object
from .geometries import GML21, GML32, is_gml_element  # also do tag registration

# All known root nodes as GML object:
FES_GML_NODES = (GM_Object, GM_Envelope, TM_Object)

__all__ = [
    "GM_Object",
    "GM_Envelope",
    "TM_Object",
    "AbstractGeometry",
    "parse_gml",
    "parse_gml_node",
    "find_gml_nodes",
    "GML21",
    "GML32",
]


def parse_gml(text: Union[str, bytes]) -> Union[FES_GML_NODES]:
    """Parse an XML <gml:...> string."""
    root_element = fromstring(text)
    return parse_gml_node(root_element)


def parse_gml_node(element: Element) -> Union[FES_GML_NODES]:
    """Parse the element"""
    if not is_gml_element(element):
        raise ExternalValueError(f"Expected GML namespace for {element.tag}")

    return tag_registry.from_child_xml(element, allowed_types=FES_GML_NODES)


def find_gml_nodes(element: Element) -> List[Element]:
    """Find all gml elements in a node"""
    result = []
    for child in element:
        # This selects all GML elements, including 2.1
        if is_gml_element(child):
            result.append(child)

    return result



"""Exceptions for the WFS operations.

Not all exception codes are listed here, only the one that
apply to the current conformance class.

See:
https://docs.opengeospatial.org/is/09-025r2/09-025r2.html#35
https://docs.opengeospatial.org/is/09-025r2/09-025r2.html#411
"""
from django.template.loader import render_to_string


class ExternalValueError(ValueError):
    """Raise a ValueError for external input.
    This helps to distinguish between internal bugs
    (e.g. unpacking values) and misformed external input.
    """


class OWSException(Exception):
    """Base class for XML based exceptions in this module."""

    status_code = 400  # Most common code in spec
    reason = None
    service = None
    version = "2.0.0"
    code = None
    text_template = None

    def __init__(self, locator, text=None, code=None):
        text = text or self.text_template.format(code=self.code, locator=locator)
        super().__init__(text)
        self.locator = locator
        self.text = text
        self.code = code or self.code

    def as_xml(self):
        return render_to_string(
            [
                f"gisserver/{self.service.lower()}/{self.version}/exception.xml",
                f"gisserver/{self.version}/exception.xml",
            ],
            {"self": self},
        )

    def __html__(self):
        return self.as_xml()


class WFSException(OWSException):
    service = "WFS"


class OperationNotSupported(WFSException):
    """WFS Method is called, but does not exist on this server."""

    status_code = 400
    reason = "Not Implemented"
    code = "OperationNotSupported"
    text_template = "Operation is not implemented."


class OperationParsingFailed(WFSException):
    """Error parsing the request to a WFS method."""

    status_code = 400
    code = "OperationParsingFailed"
    text_template = "The request could not be parsed by the server."


class OperationProcessingFailed(WFSException):
    """Error while processing the request."""

    status_code = 500
    reason = "Server processing failed"
    code = "OperationProcessingFailed"
    text_template = "The request could not be processed by the server."


class MissingParameterValue(WFSException):
    """Required parameter is missing"""

    status_code = 400
    code = "MissingParameterValue"
    text_template = "Missing required '{locator}' parameter."


class InvalidParameterValue(WFSException):
    """Unsupported choice, e.g. unsupported CRS, bad resultType."""

    status_code = 400
    code = "InvalidParameterValue"
    text_template = "Invalid value for '{locator}' parameter."


class VersionNegotiationFailed(WFSException):
    """GetCapabilities called with unsupported versions."""

    status_code = 400
    code = "VersionNegotiationFailed"
    text_template = "'ACCEPTVERSIONS' contains an invalid version number."


class NotFound(WFSException):
    """The requested ResourceId could not be found."""

    status_code = 404
    reason = "Invalid feature or property value"
    code = "NotFound"
    text_template = "Invalid feature or property value"


class OptionNotSupported(WFSException):
    status_code = 400
    reason = "Not Implemented"
    code = "OptionNotSupported"
    text_template = "Option is not implemented."


class PermissionDenied(WFSException):
    """Permission denied (custom error).

    Note this error is not part of the spec,
    but it's still useful to have access controls.
    """

    status_code = 403
    reason = "Permission Denied"
    code = "PermissionDenied"
    text_template = "You do not have permission to perform this action."



from urllib.parse import urljoin

from django.template import Library

register = Library()


@register.filter(name="urljoin")
def urljoin_(fragment, server_url):
    return urljoin(server_url, fragment)



from django.contrib.gis.db.models import GeometryField
from django.contrib.gis.db.models.functions import Transform
from typing import cast

import io

from django.conf import settings
from django.db.models import Prefetch
from django.http import HttpResponse, StreamingHttpResponse
from django.utils.html import escape

from gisserver.operations.base import WFSMethod
from gisserver.geometries import CRS
from gisserver.types import XPathMatch, XsdComplexType, XsdElement

from .results import FeatureCollection


class OutputRenderer:
    """Base class to create streaming responses.

    It receives the collected 'context' data of the WFSMethod.
    """

    #: Allow to override the maximum page size.
    #: This value can be 'math.inf' to support endless pages by default.
    max_page_size = None

    #: Define the content type for rendering the output
    content_type = "application/octet-stream"

    def __init__(
        self,
        method: WFSMethod,
        source_query,
        collection: FeatureCollection,
        output_crs: CRS,
    ):
        """
        Receive the collected data to render.

        :param method: The calling WFS Method (e.g. GetFeature class)
        :param source_query: The query that generated this output.
        :type source_query: gisserver.queries.QueryExpression
        :param collection: The collected data for rendering
        :param output_crs: The requested output projection.
        """
        self.method = method
        self.source_query = source_query
        self.collection = collection
        self.output_crs = output_crs
        self.xml_srs_name = escape(str(self.output_crs))

        # Common elements for output rendering:
        self.server_url = method.view.server_url
        self.app_xml_namespace = method.view.xml_namespace

    @classmethod
    def decorate_collection(cls, collection: FeatureCollection, output_crs, **params):
        """Perform presentation-layer logic enhancements on the queryset."""
        for sub_collection in collection.results:
            queryset = cls.decorate_queryset(
                sub_collection.feature_type,
                sub_collection.queryset,
                output_crs,
                **params,
            )
            if queryset is not None:
                sub_collection.queryset = queryset

    @classmethod
    def decorate_queryset(cls, feature_type, queryset, output_crs, **params):
        """Apply presentation layer logic to the queryset."""
        # Avoid fetching relations, fetch these within the same query,
        related = [
            Prefetch(
                xsd_element.name,
                queryset=cls.get_prefetch_queryset(xsd_element, output_crs),
            )
            for xsd_element in feature_type.xsd_type.complex_elements
        ]
        if related:
            queryset = queryset.prefetch_related(*related)

        return queryset

    @classmethod
    def get_prefetch_queryset(cls, xsd_element: XsdElement, output_crs: CRS):
        return None

    def get_response(self):
        """Render the output as streaming response."""
        stream = self.render_stream()
        if isinstance(stream, (str, bytes)):
            return HttpResponse(content=stream, content_type=self.content_type)
        else:
            stream = self._trap_exceptions(stream)
            return StreamingHttpResponse(
                streaming_content=stream, content_type=self.content_type,
            )

    def _trap_exceptions(self, stream):
        """Decorate the generator to show exceptions"""
        try:
            yield from stream
        except Exception as e:
            # Can't return 500 at this point,
            # but can still tell the client what happened.
            yield self.render_exception(e)
            raise

    def render_exception(self, exception: Exception):
        """Render the exception in a format that fits with the output."""
        if settings.DEBUG:
            return f"<!-- {exception.__class__.__name__}: {exception} -->\n"
        else:
            return f"<!-- {exception.__class__.__name__} during rendering! -->\n"

    def render_stream(self):
        """Implement this in subclasses to implement a custom output format."""
        raise NotImplementedError()


class BaseBuffer:
    """Fast buffer to write data in chunks.
    This avoids performing too many yields in the output writing.
    Especially for GeoJSON, that slows down the response times.
    """

    buffer_class = None

    def __init__(self, chunk_size=4096):
        self.data = self.buffer_class()
        self.size = 0
        self.chunk_size = chunk_size

    def is_full(self):
        return self.size >= self.chunk_size

    def write(self, value):
        if value is None:
            return
        self.size += len(value)
        self.data.write(value)

    def getvalue(self):
        return self.data.getvalue()

    def clear(self):
        self.data.seek(0)
        self.data.truncate(0)
        self.size = 0


class BytesBuffer(BaseBuffer):
    """Collect the data as bytes."""

    buffer_class = io.BytesIO

    def __bytes__(self):
        return self.getvalue()


class StringBuffer(BaseBuffer):
    """Collect the data as string"""

    buffer_class = io.StringIO

    def __str__(self):
        return self.getvalue()


def build_db_annotations(selects: dict, name_template: str, wrapper_func) -> dict:
    """Utility to build annotations for all geometry fields for an XSD type.
    This is used by various DB-optimized rendering methods.
    """
    return {
        name_template.format(name=name): wrapper_func(target)
        for name, target in selects.items()
    }


def get_db_geometry_selects(xsd_type: XsdComplexType, output_crs: CRS) -> dict:
    """Utility to generate select clauses for the geometry fields of a type."""
    return {
        xsd_element.name: _get_db_geometry_target(xsd_element, output_crs)
        for xsd_element in xsd_type.gml_elements
        if xsd_element.source is not None
    }


def _get_db_geometry_target(xsd_element: XsdElement, output_crs: CRS):
    """Wrap the selection of a geometry field in a CRS Transform if needed."""
    field = cast(GeometryField, xsd_element.source)

    target = xsd_element.model_attribute
    if field.srid != output_crs.srid:
        target = Transform(target, output_crs.srid)

    return target


def get_db_geometry_target(xpath_match: XPathMatch, output_crs: CRS):
    """Based on a resolved element, build the proper geometry field select clause."""
    field = cast(GeometryField, xpath_match.child.source)

    target = xpath_match.orm_path
    if field.srid != output_crs.srid:
        target = Transform(target, output_crs.srid)

    return target



"""All parser logic to process <fes:...> and <gml:...> tags."""

from .base import FES20 as FES20_NS
from .fes20 import Filter, function_registry as fes_function_registry
from .gml import parse_gml, GML21 as GML21_NS, GML32 as GML32_NS

parse_fes = Filter.from_string

__all__ = [
    "parse_fes",
    "parse_gml",
    "fes_function_registry",
    "FES20_NS",
    "GML21_NS",
    "GML32_NS",
]



from dataclasses import dataclass
from enum import Enum
from typing import List

from gisserver.exceptions import InvalidParameterValue
from gisserver.parsers.fes20 import ValueReference


class SortOrder(Enum):
    ASC = ""
    DESC = "-"

    # Support WFS 1 names for clients that still use this.
    A = ASC
    D = DESC

    @classmethod
    def from_string(cls, direction):
        try:
            return cls[direction]
        except KeyError:
            raise InvalidParameterValue(
                "sortby", "Expect ASC/DESC ordering direction"
            ) from None


@dataclass
class SortProperty:
    """This class name is based on the WFS spec."""

    value_reference: ValueReference
    sort_order: SortOrder = SortOrder.ASC


@dataclass
class SortBy:
    """The sortBy clause."""

    sort_properties: List[SortProperty]

    @classmethod
    def from_string(cls, value: str):
        """Construct the SortBy object from a KVP "SORTBY" parameter."""
        props = []
        for field in value.split(","):
            if "[" in field:
                raise InvalidParameterValue(
                    "sortby", "Sorting with XPath attribute selectors is not supported."
                )

            if " " in field:
                xpath, direction = field.split(" ", 1)
                props.append(
                    SortProperty(
                        value_reference=ValueReference(xpath),
                        sort_order=SortOrder.from_string(direction),
                    )
                )
            else:
                props.append(SortProperty(value_reference=ValueReference(field)))

        return cls(sort_properties=props)

    def build_ordering(self, feature_type=None) -> List[str]:
        """Build the ordering for the Django ORM call."""
        ordering = []
        for prop in self.sort_properties:
            if feature_type is not None:
                orm_field, __ = prop.value_reference.parse_xpath(feature_type)
            else:
                orm_field = prop.value_reference.xpath.replace("/", "__")

            ordering.append(f"{prop.sort_order.value}{orm_field}")
        return ordering



"""Operations that implement the WFS 2.0 spec.

Useful docs:
* http://portal.opengeospatial.org/files/?artifact_id=39967
* https://www.opengeospatial.org/standards/wfs#downloads
* http://schemas.opengis.net/wfs/2.0/
* https://mapserver.org/development/rfc/ms-rfc-105.html
* https://enonline.supermap.com/iExpress9D/API/WFS/WFS200/WFS_2.0.0_introduction.htm
"""
import math

import logging
import re
from django.core.exceptions import FieldError, ValidationError
from django.db import InternalError, ProgrammingError
from urllib.parse import urlencode

from gisserver import conf, output, queries
from gisserver.exceptions import (
    ExternalValueError,
    InvalidParameterValue,
    MissingParameterValue,
    VersionNegotiationFailed,
)
from gisserver.geometries import BoundingBox, CRS
from gisserver.parsers import fes20
from gisserver.queries import stored_query_registry

from .base import (
    OutputFormat,
    Parameter,
    WFSMethod,
    WFSTypeNamesMethod,
    UnsupportedParameter,
)

logger = logging.getLogger(__name__)

SAFE_VERSION = re.compile(r"\A[0-9.]+\Z")
RE_SAFE_FILENAME = re.compile(r"\A[A-Za-z0-9]+[A-Za-z0-9.]*")  # no dot at the start.


class GetCapabilities(WFSMethod):
    """"This operation returns map features, and available operations this WFS server supports."""

    output_formats = [OutputFormat("text/xml")]
    xml_template_name = "get_capabilities.xml"

    def get_parameters(self):
        # Not calling super, as this differs slightly (e.g. no output formats)
        return [
            Parameter(
                "service",
                required=True,
                in_capabilities=True,
                allowed_values=list(self.view.accept_operations.keys()),
            ),
            Parameter(
                # Version parameter can still be sent,
                # but not mandatory for GetCapabilities
                "version",
                allowed_values=self.view.accept_versions,
            ),
            Parameter(
                "AcceptVersions",
                in_capabilities=True,
                allowed_values=self.view.accept_versions,
                parser=self._parse_accept_versions,
            ),
            Parameter(
                "AcceptFormats",
                in_capabilities=True,
                allowed_values=self.output_formats,
                parser=self._parse_output_format,
                default=self.output_formats[0],
            ),
        ] + self.parameters

    def _parse_accept_versions(self, accept_versions) -> str:
        """Special parsing for the ACCEPTVERSIONS parameter."""
        if "VERSION" in self.view.KVP:
            # Even GetCapabilities can still receive a VERSION argument to fixate it.
            raise InvalidParameterValue(
                "AcceptVersions", "Can't provide both ACCEPTVERSIONS and VERSION"
            )

        matched_versions = set(accept_versions.split(",")).intersection(
            self.view.accept_versions
        )
        if not matched_versions:
            allowed = ", ".join(self.view.accept_versions)
            raise VersionNegotiationFailed(
                "acceptversions",
                f"'{accept_versions}' does not contain supported versions, "
                f"supported are: {allowed}.",
            )

        # Take the highest version (mapserver returns first matching)
        requested_version = sorted(matched_versions, reverse=True)[0]

        # Make sure the views+exceptions continue to operate in this version
        self.view.set_version(requested_version)

        return requested_version

    def __call__(self, **params):
        """GetCapabilities only supports XML output"""
        context = self.get_context_data(**params)
        return self.render_xml(context, **params)

    def get_context_data(self, **params):
        view = self.view

        # The 'service' is not reaad from 'params' to avoid dependency on get_parameters()
        service = view.KVP["SERVICE"]  # is WFS
        service_operations = view.accept_operations[service]
        feature_output_formats = service_operations["GetFeature"].output_formats

        return {
            "service_description": view.get_service_description(service),
            "accept_operations": {
                name: operation(view) for name, operation in service_operations.items()
            },
            "service_constraints": self.view.wfs_service_constraints,
            "filter_capabilities": self.view.wfs_filter_capabilities,
            "function_registry": fes20.function_registry,
            "accept_versions": self.view.accept_versions,
            "feature_types": self.view.get_feature_types(),
            "feature_output_formats": feature_output_formats,
            "default_max_features": self.view.max_page_size,
            "BOUNDING_BOX": conf.GISSERVER_CAPABILITIES_BOUNDING_BOX,
        }


class DescribeFeatureType(WFSTypeNamesMethod):
    """This returns an XML Schema for the provided objects.
    Each feature is exposed as an XSD definition with it's fields.
    """

    output_formats = [
        OutputFormat("XMLSCHEMA", renderer_class=output.XMLSchemaRenderer),
        # OutputFormat("text/xml", subtype="gml/3.1.1"),
    ]

    def get_context_data(self, typeNames, **params):
        if self.view.KVP.get("TYPENAMES") == "" or self.view.KVP.get("TYPENAME") == "":
            # Using TYPENAMES= does result in an error.
            raise MissingParameterValue("typeNames", f"Empty TYPENAMES parameter")
        elif typeNames is None:
            # Not given, all types are returned
            typeNames = self.all_feature_types

        return {"feature_types": typeNames}


class ListStoredQueries(WFSMethod):
    """This describes the available queries"""

    xml_template_name = "list_stored_queries.xml"

    def get_context_data(self, **params):
        return {"feature_types": self.view.get_feature_types()}


class DescribeStoredQueries(WFSMethod):
    """This describes the available queries"""

    xml_template_name = "describe_stored_queries.xml"

    parameters = [
        Parameter(
            "STOREDQUERY_ID",
            parser=lambda value: [
                stored_query_registry.resolve_query(name) for name in value.split(",")
            ],
        )
    ]

    def get_context_data(self, **params):
        queries = params["STOREDQUERY_ID"] or list(stored_query_registry)
        return {
            "feature_types": self.view.get_feature_types(),
            "stored_queries": [q.meta for q in queries],
        }


class BaseWFSGetDataMethod(WFSTypeNamesMethod):
    """Base class for GetFeature / GetPropertyValue"""

    parameters = [
        # StandardPresentationParameters
        Parameter(
            "resultType",
            in_capabilities=True,
            parser=lambda x: x.upper(),
            allowed_values=["RESULTS", "HITS"],
            default="RESULTS",
        ),
        Parameter("startIndex", parser=int, default=0),
        Parameter("count", alias="maxFeatures", parser=int),  # maxFeatures is WFS 1.x
        # outputFormat will be added by the base class.
        # StandardResolveParameters
        Parameter("resolve", allowed_values=["local"], in_capabilities=True),
        UnsupportedParameter("resolveDepth"),  # subresource settings
        UnsupportedParameter("resolveTimeout"),
        # StandardInputParameters
        Parameter("srsName", parser=CRS.from_string),
        # Projection clause parameters
        UnsupportedParameter("propertyName"),  # which fields to return
        # AdHoc Query parameters
        Parameter("bbox", parser=BoundingBox.from_string),
        Parameter(
            "filter_language",
            default=fes20.Filter.query_language,
            allowed_values=[fes20.Filter.query_language],
        ),
        Parameter("filter", parser=fes20.Filter.from_string),
        Parameter("sortBy", parser=fes20.SortBy.from_string),
        Parameter("resourceID", parser=fes20.ResourceId),
        UnsupportedParameter("aliases"),
        queries.StoredQueryParameter(),
    ]

    def get_context_data(self, resultType, **params):  # noqa: C901
        query = self.get_query(**params)
        query.check_permissions(self.view.request)

        try:
            if resultType == "HITS":
                collection = self.get_hits(query)
            elif resultType == "RESULTS":
                # Validate StandardPresentationParameters
                collection = self.get_paginated_results(query, **params)
            else:
                raise NotImplementedError()
        except ExternalValueError as e:
            # Bad input data
            self._log_filter_error(query, logging.ERROR, e)
            raise InvalidParameterValue(
                self._get_locator(**params), f"Invalid filter query: {e}",
            ) from e
        except ValidationError as e:
            # Bad input data
            self._log_filter_error(query, logging.ERROR, e)
            msg = "\n".join(map(str, e.messages))
            raise InvalidParameterValue(
                self._get_locator(**params), f"Invalid filter query: {msg}",
            ) from e
        except FieldError as e:
            # e.g. doing a LIKE on a foreign key, or requesting an unknown field.
            self._log_filter_error(query, logging.ERROR, e)
            raise InvalidParameterValue(
                self._get_locator(**params), f"Internal error when processing filter",
            ) from e
        except (InternalError, ProgrammingError) as e:
            # e.g. comparing datetime against integer
            logger.error("WFS request failed: %s\nParams: %r", e, params)
            msg = str(e)
            locator = (
                "srsName" if "Cannot find SRID" in msg else self._get_locator(**params)
            )
            raise InvalidParameterValue(locator, f"Invalid request: {msg}") from e
        except (TypeError, ValueError) as e:
            # TypeError/ValueError could reference a datatype mismatch in an
            # ORM query, but it could also be an internal bug. In most cases,
            # this is already caught by XsdElement.validate_comparison().
            if self._is_orm_error(e):
                raise InvalidParameterValue(
                    self._get_locator(**params), f"Invalid filter query: {e}",
                ) from e
            raise

        output_crs = params["srsName"]
        if not output_crs and collection.results:
            output_crs = collection.results[0].feature_type.crs

        # These become init kwargs for the selected OutputRenderer class:
        return {
            "source_query": query,
            "collection": collection,
            "output_crs": output_crs,
        }

    def get_query(self, **params) -> queries.QueryExpression:
        """Create the query object that will process this request"""
        if params["STOREDQUERY_ID"]:
            # The 'StoredQueryParameter' already parses the input into a complete object.
            # When it's not provided, the regular Adhoc-query will be created from the KVP request.
            query = params["STOREDQUERY_ID"]
        else:
            query = queries.AdhocQuery.from_kvp_request(**params)

        # TODO: pass this in a cleaner way?
        query.bind(
            all_feature_types=self.all_feature_types_by_name,  # for GetFeatureById
            value_reference=params.get("valueReference"),  # for GetPropertyValue
        )

        return query

    def get_hits(self, query: queries.QueryExpression) -> output.FeatureCollection:
        """Return the number of hits"""
        return query.get_hits()

    def get_results(
        self, query: queries.QueryExpression, start, count
    ) -> output.FeatureCollection:
        """Return the actual results"""
        return query.get_results(start, count=count)

    def get_paginated_results(
        self, query: queries.QueryExpression, outputFormat, **params
    ) -> output.FeatureCollection:
        """Handle pagination settings."""
        start = max(0, params["startIndex"])

        # outputFormat.max_page_size can be math.inf to enable endless scrolling.
        # this only works when the COUNT parameter is not given.
        max_page_size = outputFormat.max_page_size or self.view.max_page_size
        page_size = min(max_page_size, params["count"] or max_page_size)
        stop = start + page_size

        # Perform query
        collection = self.get_results(query, start=start, count=page_size)

        # Allow presentation-layer to add extra logic.
        if outputFormat.renderer_class is not None:
            output_crs = params["srsName"]
            if not output_crs and collection.results:
                output_crs = collection.results[0].feature_type.crs

            outputFormat.renderer_class.decorate_collection(
                collection, output_crs, **params
            )

        if stop != math.inf:
            if start > 0:
                collection.previous = self._replace_url_params(
                    STARTINDEX=max(0, start - page_size), COUNT=page_size,
                )
            if stop < collection.number_matched:
                # TODO: fix this when returning multiple typeNames:
                collection.next = self._replace_url_params(
                    STARTINDEX=start + page_size, COUNT=page_size
                )

        return collection

    def _replace_url_params(self, **updates) -> str:
        """Replace a query parameter in the URL"""
        new_params = self.view.KVP.copy()
        new_params.update(updates)
        return f"{self.view.server_url}?{urlencode(new_params)}"

    def _is_orm_error(self, exception: Exception):
        traceback = exception.__traceback__
        while traceback.tb_next is not None:
            traceback = traceback.tb_next
            if "/django/db/models/query" in traceback.tb_frame.f_code.co_filename:
                return True
        return False

    def _log_filter_error(self, query, level, exc):
        """Report a filtering parsing error in the logging"""
        filter = getattr(query, "filter", None)  # AdhocQuery only
        fes_xml = filter.source if filter is not None else "(not provided)"
        try:
            sql = exc.__cause__.cursor.query.decode()
        except AttributeError:
            logger.log(level, "WFS query failed: %s\nFilter:\n%s", exc, fes_xml)
        else:
            logger.log(
                level,
                "WFS query failed: %s\nSQL Query: %s\n\nFilter:\n%s",
                exc,
                sql,
                fes_xml,
            )

    def _get_locator(self, **params):
        """Tell which field is likely causing the query error"""
        if params["resourceID"]:
            return "resourceId"
        elif params["STOREDQUERY_ID"]:
            return "STOREDQUERY_ID"
        else:
            return "filter"


class GetFeature(BaseWFSGetDataMethod):
    """This returns all properties of the feature.

    Various query parameters allow limiting the data.
    """

    output_formats = [
        OutputFormat(
            # Needed for cite compliance tests
            "application/gml+xml",
            version="3.2",
            renderer_class=output.gml32_renderer,
        ),
        OutputFormat(
            "text/xml", subtype="gml/3.2.1", renderer_class=output.gml32_renderer
        ),
        # OutputFormat("gml"),
        OutputFormat(
            "application/json",
            subtype="geojson",
            charset="utf-8",
            renderer_class=output.geojson_renderer,
        ),
        OutputFormat(
            "text/csv",
            subtype="csv",
            charset="utf-8",
            renderer_class=output.csv_renderer,
        ),
        # OutputFormat("shapezip"),
        # OutputFormat("application/zip"),
    ]


class GetPropertyValue(BaseWFSGetDataMethod):
    """This returns a limited set of properties of the feature.
    It works almost identical to GetFeature, except that it returns a single field.
    """

    output_formats = [
        OutputFormat(
            "text/xml", subtype="gml/3.2", renderer_class=output.gml32_value_renderer
        ),
    ]

    parameters = BaseWFSGetDataMethod.parameters + [
        Parameter("valueReference", required=True, parser=fes20.ValueReference)
    ]

    def get_context_data(self, **params):
        # Pass valueReference to output format.
        # NOTE: The AdhocQuery object also performs internal processing,
        # so the query performs a "SELECT id, <fieldname>" as well.
        # In the WFS-spec, the valueReference is only a presentation layer change.
        context = super().get_context_data(**params)
        context["value_reference"] = params["valueReference"]
        return context



"""Internal XSD type definitions.

These types are the internal definition on which all output is generated.
It's constructed from the model metadata by the `FeatureType` / `FeatureField`
classes. Custom field types could also generate these field types.
"""
import re
from dataclasses import dataclass
from enum import Enum

from django.core.exceptions import ValidationError
from django.db.models import Q
from typing import List, Optional, Type

from django.contrib.gis.db.models import GeometryField
from django.db import models
from django.utils.functional import cached_property

from gisserver.geometries import CRS, WGS84  # noqa, for backwards compatibility

__all__ = [
    "XsdAnyType",
    "XsdTypes",
    "XsdComplexType",
    "strip_namespace_prefix",
]


RE_XPATH_ATTR = re.compile(r"\[[^\]]+\]$")
RE_NON_NAME = re.compile(r"[^a-zA-Z0-9_/]")


class XsdAnyType:
    """Base class for all types used in the XML definition"""

    prefix = None
    is_complex_type = False

    def __str__(self):
        raise NotImplementedError()

    def with_prefix(self, prefix="xs"):
        raise NotImplementedError()


class XsdTypes(XsdAnyType, Enum):
    """Brief enumeration of basic XSD-types.

    The default namespace is the "xs:" (XMLSchema).
    Based on https://www.w3.org/TR/xmlschema-2/#built-in-datatypes
    """

    anyType = "anyType"  # Needs to be anyType, as "xsd:any" is an element, not a type.
    string = "string"
    boolean = "boolean"
    decimal = "decimal"  # the base type for all numbers too.
    integer = "integer"
    float = "float"
    double = "double"
    time = "time"
    date = "date"
    dateTime = "dateTime"
    anyURI = "anyURI"

    # Less common, but useful nonetheless:
    duration = "duration"
    nonNegativeInteger = "nonNegativeInteger"
    gYear = "gYear"
    hexBinary = "hexBinary"
    base64Binary = "base64Binary"
    token = "token"
    language = "language"

    # Types that contain a GML value as member:
    gmlGeometryPropertyType = "gml:GeometryPropertyType"
    gmlPointPropertyType = "gml:PointPropertyType"
    gmlCurvePropertyType = "gml:CurvePropertyType"  # curve is base for LineString
    gmlSurfacePropertyType = "gml:SurfacePropertyType"  # GML2 had PolygonPropertyType
    gmlMultiSurfacePropertyType = "gml:MultiSurfacePropertyType"
    gmlMultiPointPropertyType = "gml:MultiPointPropertyType"
    gmlMultiCurvePropertyType = "gml:MultiCurvePropertyType"
    gmlMultiGeometryPropertyType = "gml:MultiGeometryPropertyType"

    #: A direct geometry value
    gmlAbstractGeometryType = "gml:AbstractGeometryType"

    #: A feature that has an gml:name and gml:boundedBy as posible child element.
    gmlAbstractFeatureType = "gml:AbstractFeatureType"

    def __str__(self):
        return self.value

    @property
    def prefix(self) -> Optional[str]:
        colon = self.value.find(":")
        return self.value[:colon] if colon else None

    def with_prefix(self, prefix="xs"):
        if ":" in self.value:
            return self.value
        else:
            return f"{prefix}:{self.value}"


@dataclass(frozen=True)
class XsdElement:
    """Declare an XSD element"""

    name: str
    type: XsdAnyType  # Both XsdComplexType and XsdType are allowed
    nillable: Optional[bool] = None
    min_occurs: Optional[int] = None
    max_occurs: Optional[int] = None
    source: Optional[models.Field] = None

    #: Which field to read from the model to get the value
    model_attribute: Optional[str] = None

    def __post_init__(self):
        if self.model_attribute is None:
            object.__setattr__(self, "model_attribute", self.name)

    @cached_property
    def is_gml(self):
        return isinstance(self.source, GeometryField) or self.type.prefix == "gml"

    @cached_property
    def as_xml(self):
        attributes = [f'name="{self.name}" type="{self.type}"']
        if self.min_occurs is not None:
            attributes.append(f'minOccurs="{self.min_occurs}"')
        if self.max_occurs is not None:
            attributes.append(f'maxOccurs="{self.max_occurs}"')
        if self.nillable:
            str_bool = "true" if self.nillable else "false"
            attributes.append(f'nillable="{str_bool}"')

        return "<element {} />".format(" ".join(attributes))

    def __str__(self):
        return self.as_xml

    def get_value(self, instance: models.Model):
        """Provide the value for the """
        # For foreign keys, it's not possible to use the model value,
        # as that would conflict with the field type in the XSD schema.
        try:
            return getattr(instance, self.model_attribute)
        except AttributeError:
            # E.g. Django foreign keys that point to a non-existing member.
            return None

    def validate_comparison(self, value):
        """Validate whether the input value can be used in a comparison.
        This avoids comparing a database DATETIME object to an integer.
        """
        if self.source is not None:
            # Not calling self.source.validate() as that checks for allowed choices,
            # which shouldn't be checked against for a filter query.
            try:
                self.source.get_prep_value(value)
            except (ValueError, TypeError) as e:
                raise ValidationError(str(e)) from e


@dataclass(frozen=True)
class XsdComplexType(XsdAnyType):
    """Define an <xsd:complexType> that represents a whole class definition.

    By default, The type is declared as subclass of <gml:AbstractFeatureType>,
    which allows child elements like <gml:name> and <gml:boundedBy>.
    """

    name: str
    elements: List[XsdElement]
    base: XsdTypes = XsdTypes.gmlAbstractFeatureType
    source: Optional[Type[models.Model]] = None

    def __str__(self):
        return f"{self.prefix}:{self.name}"

    @property
    def is_complex_type(self):
        return True

    @property
    def prefix(self):
        # mimic API of XsdTypes
        return "app"

    def with_prefix(self, prefix="xs"):
        # mimic API of XsdTypes
        return str(self)

    @cached_property
    def gml_elements(self) -> List[XsdElement]:
        """Shortcut to get all geometry elements"""
        return [e for e in self.elements if e.is_gml]

    @cached_property
    def complex_elements(self) -> List[XsdElement]:
        """Shortcut to get all elements with a complex type"""
        return [e for e in self.elements if e.type.is_complex_type]

    def resolve_element_path(self, xpath: str) -> Optional[List[XsdElement]]:
        """Resolve an xpath reference to the actual node.
        This returns the list of all levels if a match was found.
        """
        try:
            pos = xpath.rindex("/")
            node_name = xpath[:pos]
        except ValueError:
            node_name = xpath
            pos = 0

        # Strip current app namespace. Note this should actually compare the
        # xmlns URI's, but this will suffice for now. The ElementTree parser
        # doesn't provide access to 'xmlns' definitions on the element (or it's
        # parents), so a tag like this is essentially not parsable for us:
        # <ValueReference xmlns:tns="http://example.org/gisserver">tns:fieldname</ValueReference>
        node_name = strip_namespace_prefix(node_name)

        # Strip any [@attr=..] conditions
        node_name = RE_XPATH_ATTR.sub("", node_name)

        for element in self.elements:
            if element.name == node_name:
                if pos:
                    if not element.type.is_complex_type:
                        return None
                    else:
                        child_path = element.type.resolve_element_path(xpath[pos + 1 :])
                        if child_path is None:
                            return None
                        else:
                            return [element] + child_path
                else:
                    return [element]

        return None


def strip_namespace_prefix(value: str):
    """Remove the namespace prefix from an element."""
    try:
        ns_pos = value.index(":")
        return value[ns_pos + 1 :]
    except ValueError:
        return value


class XPathMatch:
    """Wrapper class to provide XPath results."""

    #: The matched element, with all it's parents.
    elements: List[XsdElement]

    #: The source XPath query
    query: str

    def __init__(self, elements: List[XsdElement], query: str):
        self.elements = elements
        self.query = query

    def __iter__(self):
        return iter(self.elements)

    def __getitem__(self, item) -> XsdElement:
        return self.elements[item]

    def __repr__(self):
        return f"XPathMatch(elements={self.elements!r}, query={self.query!r})"

    @property
    def child(self) -> XsdElement:
        """Return only the final element"""
        return self.elements[-1]

    @property
    def orm_path(self) -> str:
        """Tell which ORM path should be targetted."""
        return "__".join(xsd_element.model_attribute for xsd_element in self.elements)

    @property
    def orm_filters(self) -> Optional[Q]:
        """Tell which additional filters are needed (due to [@attr=..] syntax)."""
        if RE_NON_NAME.match(self.query):
            # If there is an element[@attr=...]/field tag,
            # the build_...() logic should return a Q() object.
            raise NotImplementedError(
                f"Complex XPath queries are not supported yet: {self.query}"
            )

        return None



from .expressions import ValueReference
from .filters import Filter
from .functions import function_registry
from .identifiers import ResourceId
from .query import CompiledQuery
from .sorting import SortBy

__all__ = [
    "Filter",
    "CompiledQuery",
    "ValueReference",
    "ResourceId",
    "SortBy",
    "function_registry",
]



import re
from datetime import date, datetime, time
from decimal import Decimal as D
from functools import wraps
from typing import List, Optional, Tuple
from xml.etree.ElementTree import Element, QName

from gisserver.exceptions import ExternalValueError

RE_FLOAT = re.compile(r"\A[0-9]+(\.[0-9]+)\Z")


def expect_tag(namespace, *tag_names, leaf=False):
    """Validate whether a given tag is need"""
    valid_tags = set(str(QName(namespace, name)) for name in tag_names)
    expect0 = str(QName(namespace, tag_names[0]))

    def _wrapper(func):
        @wraps(func)
        def _from_xml_expect(cls, element, *args, **kwargs):
            if element.tag not in valid_tags:
                raise ExternalValueError(
                    f"{cls.__name__}.{func.__name__}(element) expects an <{expect0}> node, "
                    f"got <{element.tag}>"
                )
            if leaf and len(element):
                raise ExternalValueError(
                    f"{element.tag} elements should not have child nodes."
                )

            return func(cls, element, *args, **kwargs)

        return _from_xml_expect

    return _wrapper


def get_child(root, namespace, localname) -> Element:
    """Find the element using a fully qualified name."""
    return root.find(QName(namespace, localname).text)


def get_children(root, namespace, localname) -> List[Element]:
    """Find the element using a fully qualified name."""
    return root.findall(QName(namespace, localname).text)


def get_attribute(element: Element, name) -> str:
    """Resolve an attribute, raise an error when it's missing."""
    try:
        return element.attrib[name]
    except KeyError:
        raise ExternalValueError(
            f"Element {element.tag} misses required attribute '{name}'"
        ) from None


def split_ns(tag_name: str) -> Tuple[Optional[str], str]:
    """Split the element tag into the namespace and local name.
    The stdlib etree doesn't have the properties for this (lxml does).
    """
    if tag_name.startswith("{"):
        end = tag_name.index("}")
        return tag_name[1:end], tag_name[end + 1 :]
    else:
        return None, tag_name


def auto_cast(value: str):
    """Automatically cast a value to a scalar."""
    if value.isdigit():
        return int(value)
    elif RE_FLOAT.match(value):
        return D(value)
    elif "T" in value:
        try:
            return parse_iso_datetime(value)
        except ValueError:
            pass

    return value


def xsd_cast(value: str, type: str):
    if type == "xs:date":
        return date.fromisoformat(value)
    elif type == "xs:dateTime":
        return parse_iso_datetime(value)
    elif type == "xs:time":
        return time.fromisoformat(value)
    elif type == "xs:string":
        return value
    elif type in ("xs:int", "xs:integer", "xs:long"):
        return int(value)
    elif type in ("xs:float", "xs:double", "xs:decimal"):
        return D(value)
    else:
        raise NotImplementedError(f'<fes:Literal type="{type}"> is not implemented.')


def parse_iso_datetime(value) -> datetime:
    try:
        return datetime.fromisoformat(value)  # Python 3.7+
    except AttributeError:
        return datetime.strptime(value, "%Y-%m-%dT%H:%M:%S")
